@@ -10,15 +10,16 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 #
-# Copyright 2014 Nextdoor.com, Inc
+# Copyright 2018 Nextdoor.com, Inc
 
-"""AWS SQS Actors"""
+"""
+:mod:`kingpin.actors.aws.sqs`
+^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
+"""
 
 import logging
 import re
-import time
 
-from concurrent import futures
 from tornado import concurrent
 from tornado import gen
 from tornado import ioloop
@@ -27,9 +28,11 @@ import boto.sqs.queue
 import mock
 
 from kingpin import utils
-from kingpin.actors import base
 from kingpin.actors import exceptions
+from kingpin.actors.aws import base
 from kingpin.actors.aws import settings as aws_settings
+from kingpin.actors.utils import dry
+from kingpin.constants import REQUIRED
 
 log = logging.getLogger(__name__)
 
@@ -40,26 +43,30 @@ __author__ = 'Mikhail Simin <mikhail@nextdoor.com>'
 # decorator. We would like this to be a class variable so its shared
 # across RightScale objects, but we see testing IO errors when we
 # do this.
-EXECUTOR = futures.ThreadPoolExecutor(10)
+EXECUTOR = concurrent.futures.ThreadPoolExecutor(10)
 
 
-class SQSActorException(Exception):
-
-    """Raised by SQS Actor for any exception."""
-
-
-class SQSQueueNotFoundException(SQSActorException):
+class QueueNotFound(exceptions.RecoverableActorFailure):
 
     """Raised by SQS Actor when a needed queue is not found."""
 
 
-class SQSBaseActor(base.BaseActor):
+class QueueDeletionFailed(exceptions.RecoverableActorFailure):
+
+    """Raised if Boto fails to delete an SQS queue.
+
+    http://boto.readthedocs.org/en/latest/ref/
+        sqs.html#boto.sqs.connection.SQSConnection.delete_queue
+    """
+
+
+class SQSBaseActor(base.AWSBaseActor):
 
     # This actor should not be instantiated, but unit testing requires that
     # it's all options are defined properly here.
     all_options = {
-        'name': (str, None, 'Queue name to do nothing with.'),
-        'region': (str, None, 'AWS region name like us-west-2')
+        'name': (str, REQUIRED, 'Queue name to do nothing with.'),
+        'region': (str, REQUIRED, 'AWS region (or zone) name like us-west-2')
     }
 
     # Get references to existing objects that are used by the
@@ -67,46 +74,7 @@ class SQSBaseActor(base.BaseActor):
     ioloop = ioloop.IOLoop.current()
     executor = EXECUTOR
 
-    def __init__(self, *args, **kwargs):
-        """Create the connection object."""
-        super(SQSBaseActor, self).__init__(*args, **kwargs)
-
-        region = self._get_region(self.option('region'))
-
-        if not (aws_settings.AWS_ACCESS_KEY_ID and
-                aws_settings.AWS_SECRET_ACCESS_KEY):
-            raise exceptions.InvalidCredentials(
-                'AWS settings imported but not all credentials are supplied. '
-                'AWS_ACCESS_KEY_ID: %s, AWS_SECRET_ACCESS_KEY: %s' % (
-                    aws_settings.AWS_ACCESS_KEY_ID,
-                    aws_settings.AWS_SECRET_ACCESS_KEY))
-
-        self.conn = boto.sqs.connection.SQSConnection(
-            aws_settings.AWS_ACCESS_KEY_ID,
-            aws_settings.AWS_SECRET_ACCESS_KEY,
-            region=region)
-
-    def _get_region(self, region):
-        """Return 'region' object used in SQSConnection
-
-        Args:
-            region: string - AWS region name, like us-west-2
-        Returns:
-            RegionInfo object from boto.sqs
-        """
-
-        all_regions = boto.sqs.regions()
-        match = [r for r in all_regions if r.name == region]
-
-        if len(match) != 1:
-            raise exceptions.UnrecoverableActionFailure((
-                'Expected to find exactly 1 region named %s. '
-                'Found: %s') % (region, match))
-
-        return match[0]
-
-    @concurrent.run_on_executor
-    @utils.exception_logger
+    @gen.coroutine
     def _fetch_queues(self, pattern):
         """Searches SQS for all queues with a matching name pattern.
 
@@ -116,22 +84,47 @@ class SQSBaseActor(base.BaseActor):
         Returns:
             Array of matched queues, even if empty.
         """
-        queues = self.conn.get_all_queues()
+        queues = yield self.api_call(self.sqs_conn.get_all_queues)
         match_queues = [q for q in queues if re.search(pattern, q.name)]
-        return match_queues
+        raise gen.Return(match_queues)
 
 
 class Create(SQSBaseActor):
 
-    """Creates a new SQS Queue."""
+    """Creates a new SQS queue with the specified name
+
+    **Options**
+
+    :name:
+      (str) The name of the queue to create
+
+    :region:
+      (str) AWS region (or zone) string, like 'us-west-2'
+
+    **Examples**
+
+    .. code-block:: json
+
+       { "actor": "aws.sqs.Create",
+         "desc": "Create queue named async-tasks",
+         "options": {
+           "name": "async-tasks",
+           "region": "us-east-1",
+         }
+       }
+
+    **Dry Mode**
+
+    Will not create any queue, or even contact SQS. Will create a mock.Mock
+    object and exit with success.
+    """
 
     all_options = {
-        'name': (str, None, 'Name or pattern for SQS queues.'),
-        'region': (str, None, 'AWS region for SQS, such as us-west-2')
+        'name': (str, REQUIRED, 'Name or pattern for SQS queues.'),
+        'region': (str, REQUIRED, 'AWS region (or zone), such as us-west-2')
     }
 
-    @concurrent.run_on_executor
-    @utils.exception_logger
+    @gen.coroutine
     def _create_queue(self, name):
         """Create an SQS queue with the specified name.
 
@@ -146,20 +139,20 @@ class Create(SQSBaseActor):
         """
         if not self._dry:
             self.log.info('Creating a new queue: %s' % name)
-            new_queue = self.conn.create_queue(name)
+            new_queue = yield self.api_call(self.sqs_conn.create_queue, name)
         else:
             self.log.info('Would create a new queue: %s' % name)
             new_queue = mock.Mock(name=name)
 
         self.log.debug('Returning queue object: %s' % new_queue)
-        return new_queue
+        raise gen.Return(new_queue)
 
     @gen.coroutine
     def _execute(self):
         """Executes an actor and yields the results when its finished.
 
         Raises:
-            gen.Return(True)
+            gen.Return()
         """
         q = yield self._create_queue(name=self.option('name'))
 
@@ -168,47 +161,95 @@ class Create(SQSBaseActor):
         elif self._dry:
             self.log.info('Fake Queue: %s' % q)
         else:
-            raise exceptions.UnrecoverableActionFailure(
+            raise exceptions.UnrecoverableActorFailure(
                 'All hell broke loose: %s' % q)
 
-        raise gen.Return(True)
+        raise gen.Return()
 
 
 class Delete(SQSBaseActor):
 
-    """Deletes an existing SQS Queue."""
+    """Deletes the SQS queues
+
+    Note: **even if it`s not empty**
+
+    **Options**
+
+    :name:
+      (str) The name of the queue to destroy
+
+    :region:
+      (str) AWS region (or zone) string, like 'us-west-2'
+
+    :idempotent:
+      (bool) Will not raise errors if no matching queues are found.
+      (default: False)
+
+    **Examples**
+
+    .. code-block:: json
+
+       { "actor": "aws.sqs.Delete",
+         "desc": "Delete queue async-tasks",
+         "options": {
+           "name": "async-tasks",
+           "region": "us-east-1"
+         }
+       }
+
+    .. code-block:: json
+
+       { "actor": "aws.sqs.Delete",
+         "desc": "Delete queues with 1234 in the name",
+         "options": {
+           "name": "1234",
+           "region": "us-east-1"
+         }
+       }
+
+    **Dry Mode**
+
+    Will find the specified queue, but will have a noop regarding its deletion.
+    Dry mode will fail if no queues are found, and idempotent flag is set to
+    False.
+    """
 
     all_options = {
-        'name': (str, None, 'Name or pattern for SQS queues.'),
-        'region': (str, None, 'AWS region for SQS, such as us-west-2'),
+        'name': (str, REQUIRED, 'Name or pattern for SQS queues.'),
+        'region': (str, REQUIRED, 'AWS region (or zone), such as us-west-2'),
         'idempotent': (bool, False, 'Continue if queues are already deleted.')
     }
 
-    @concurrent.run_on_executor
-    @utils.exception_logger
+    @gen.coroutine
+    @dry('Would delete the queue: {queue}')
     def _delete_queue(self, queue):
         """Delete the provided queue.
 
-        Raises UnrecoverableActionFailure if fail to delete it.
+        Raises RecoverableActorFailure if fail to delete it.
 
-        Returns True if successful in deletion, or is Dry run.
+        Returns:
+          True if successful in deletion, or is Dry run.
+
+        Raises:
+          QueueDeletionFailed if queue deletion failed.
         """
-        if not self._dry:
-            self.log.info('Deleting Queue: %s...' % queue.url)
-            ok = self.conn.delete_queue(queue)
-        else:
-            self.log.info('Would delete the queue: %s' % queue.url)
-            ok = True
+        self.log.info('Deleting Queue: %s...' % queue.url)
+        ok = yield self.api_call(self.sqs_conn.delete_queue, queue)
 
-        return ok
+        # Raise an exception if the tasks failed
+        if not ok:
+            raise QueueDeletionFailed('Failed to delete "%s"' % queue.url)
+
+        raise gen.Return(ok)
 
     @gen.coroutine
-    @utils.retry(SQSQueueNotFoundException, delay=aws_settings.SQS_RETRY_DELAY)
+    @utils.retry(QueueNotFound, delay=aws_settings.SQS_RETRY_DELAY)
     def _execute(self):
         """Executes an actor and yields the results when its finished.
 
         Raises:
-            gen.Return(True)
+            gen.Return()
+            QueueNotFound()
         """
         pattern = self.option('name')
         matched_queues = yield self._fetch_queues(pattern=pattern)
@@ -217,32 +258,68 @@ class Delete(SQSBaseActor):
                                not self.option('idempotent'))
 
         if not_found_condition:
-            raise SQSQueueNotFoundException(
+            raise QueueNotFound(
                 'No queues with pattern "%s" found.' % pattern)
 
         self.log.info('Deleting SQS Queues: %s' % matched_queues)
 
         tasks = []
         for q in matched_queues:
-            tasks.append(self._delete_queue(q))
+            tasks.append(self._delete_queue(queue=q))
+        yield tasks
 
-        result = yield tasks
-
-        raise gen.Return(all(result))
+        raise gen.Return()
 
 
 class WaitUntilEmpty(SQSBaseActor):
 
-    """Waits for one or more SQS Queues to become empty."""
+    """Wait indefinitely until for SQS queues to become empty
+
+    This actor will loop infinitely as long as the count of messages in at
+    least one queue is greater than zero. SQS does not guarantee exact count,
+    so this can return a stale value if the number of messages in the queue
+    changes rapidly.
+
+
+    **Options**
+
+    :name:
+      (str) The name or regex pattern of the queues to operate on
+
+    :region:
+      (str) AWS region (or zone) string, like 'us-west-2'
+
+    :required:
+      (bool) Fail if no matching queues are found.
+      (default: False)
+
+    **Examples**
+
+    .. code-block:: json
+
+       { "actor": "aws.sqs.WaitUntilEmpty",
+         "desc": "Wait until release-0025a* queues are empty",
+         "options": {
+           "name": "release-0025a",
+           "region": "us-east-1",
+           "required": true
+         }
+       }
+
+    **Dry Mode**
+
+    This actor performs the finding of the queue, but will pretend that the
+    count is 0 and return success. Will fail even in dry mode if `required`
+    option is set to True and no queues with the name pattern are found.
+    """
 
     all_options = {
-        'name': (str, None, 'Name or pattern for SQS queues.'),
-        'region': (str, None, 'AWS region for SQS, such as us-west-2'),
+        'name': (str, REQUIRED, 'Name or pattern for SQS queues.'),
+        'region': (str, REQUIRED, 'AWS region (or zone), such as us-west-2'),
         'required': (bool, False, 'At least 1 queue must be found.')
     }
 
-    @concurrent.run_on_executor
-    @utils.exception_logger
+    @gen.coroutine
     def _wait(self, queue, sleep=3):
         """Sleeps until an SQS Queue has emptied out.
 
@@ -258,7 +335,11 @@ class WaitUntilEmpty(SQSBaseActor):
         while True:
             if not self._dry:
                 self.log.debug('Counting %s' % queue.url)
-                count = queue.count()
+                visible = yield self.api_call(queue.count)
+                attr = 'ApproximateNumberOfMessagesNotVisible'
+                invisible = yield self.api_call(queue.get_attributes, attr)
+                invisible_int = int(invisible[attr])
+                count = visible + invisible_int
             else:
                 self.log.info('Pretending that count is 0 for %s' % queue.url)
                 count = 0
@@ -266,25 +347,25 @@ class WaitUntilEmpty(SQSBaseActor):
             self.log.debug('Queue has %s messages in it.' % count)
             if count > 0:
                 self.log.info('Waiting on %s to become empty...' % queue.name)
-                time.sleep(sleep)
+                yield utils.tornado_sleep(sleep)
             else:
                 self.log.debug('Queue is empty!')
                 break
 
-        return True
+        raise gen.Return(True)
 
     @gen.coroutine
     def _execute(self):
         """Executes an actor and yields the results when its finished.
 
-        raises: gen.Return(True)
+        raises: gen.Return()
         """
         pattern = self.option('name')
         matched_queues = yield self._fetch_queues(pattern)
 
         # Note: this does not check for dry mode.
         if self.option('required') and not matched_queues:
-            raise exceptions.ActorException(
+            raise QueueNotFound(
                 'No queues like "%s" were found!' % pattern)
 
         self.log.info('Waiting for "%s" queues to become empty.' %
@@ -299,4 +380,4 @@ class WaitUntilEmpty(SQSBaseActor):
         yield sleepers
         self.log.info('All queues report empty.')
 
-        raise gen.Return(True)
+        raise gen.Return()
