@@ -10,7 +10,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 #
-# Copyright 2014 Nextdoor.com, Inc
+# Copyright 2018 Nextdoor.com, Inc
 
 import jsonschema
 
@@ -19,9 +19,9 @@ from kingpin import exceptions
 __author__ = 'Matt Wise <matt@nextdoor.com>'
 
 
-SCHEMA_1_0 = {
-    'type': 'object',
-    'required': ['actor', 'desc', 'options'],
+ACTOR_SCHEMA = {
+    'type': ['object'],
+    'required': ['actor'],
     'additionalProperties': False,
     'properties': {
         'desc': {'type': 'string'},
@@ -51,10 +51,27 @@ SCHEMA_1_0 = {
         },
 
         # Not required. In code, will default to False.
-        'warn_on_failure': {'type': 'boolean'},
+        'warn_on_failure': {'type': ['boolean', 'string']},
+
+        # Not required. In code, will default to <actor>.default_timeout
+        'timeout': {'type': ['string', 'integer', 'number']},
+
+        # Optional conditional to indicate to skip this actor.
+        'condition': {'type': ['boolean', 'string'], 'default': True},
     }
 }
 
+SCHEMA_1_0 = {
+    'definitions': {
+        'actor': ACTOR_SCHEMA
+    },
+
+    'anyOf': [
+        {'$ref': '#/definitions/actor'},
+        {'type': 'array', 'items': {'$ref': '#/definitions/actor'}}
+    ]
+}
+
 
 def validate(config):
     """Validates the JSON against our schemas.
@@ -73,4 +90,4 @@ def validate(config):
     try:
         return jsonschema.validate(config, SCHEMA_1_0)
     except jsonschema.exceptions.ValidationError as e:
-        raise exceptions.InvalidJSON(e)
+        raise exceptions.InvalidScript(e)
